@@ -25,13 +25,11 @@
 /*declarations of structures*/
 #include "structs.h"
 
-typedef GelETree *(*dictfunc)(GelCtx *ctx, GelETree * * /*arguments*/,int * /*exception*/);
-
 /*return current context number (0 is global, -1 is uninitialized)*/
 int d_curcontext(void);
 
 /*make builtin function and return it*/
-GelEFunc * d_makebifunc(GelToken *id, dictfunc f, int nargs);
+GelEFunc * d_makebifunc(GelToken *id, GelBIFunction f, int nargs);
 
 /*make a user function and return it*/
 GelEFunc * d_makeufunc (GelToken *id, GelETree *value, GSList *argnames, int nargs,
@@ -40,18 +38,18 @@ GelEFunc * d_makeufunc (GelToken *id, GelETree *value, GSList *argnames, int nar
 /*make a variable function and return in*/
 GelEFunc * d_makevfunc(GelToken *id, GelETree *value);
 
-/*make a user function and return it*/
+/*make a reference function and return it*/
 GelEFunc * d_makereffunc(GelToken *id, GelEFunc *ref);
 
 /*copy a function*/
 GelEFunc *d_copyfunc(GelEFunc *o);
 
 /*make a real function from a fake*/
-GelEFunc * d_makerealfunc(GelEFunc *o,GelToken *id, int use);
+GelEFunc * d_makerealfunc(GelEFunc *o,GelToken *id, gboolean use);
 
 /*make real func and replace o with it, without changing o's context or id*/
 /*if use is set, we USE the original function, NULLing approriately*/
-void d_setrealfunc(GelEFunc *n,GelEFunc *fake, int use);
+void d_setrealfunc(GelEFunc *n,GelEFunc *fake, gboolean use);
 
 void d_initcontext(void);
 
@@ -62,7 +60,7 @@ GelEFunc * d_addfunc_global (GelEFunc *func);
 
 /*set value of an existing function (in local context), used for arguments
   WARNING, does not free the memory allocated by previous value!*/
-int d_setvalue(GelToken *id,GelETree *value);
+gboolean d_setvalue (GelToken *id,GelETree *value);
 
 /*this will work right in all situations*/
 void d_set_value(GelEFunc *n,GelETree *value);
@@ -82,7 +80,7 @@ GelEFunc * d_lookup_only_global (GelToken *id);
 
 GelToken * d_intern (const char *id);
 
-int d_delete(GelToken *id);
+gboolean d_delete(GelToken *id);
 
 /*clear all context dictionaries and pop out all the contexts except
   the global one
@@ -95,10 +93,10 @@ void d_freedict(GSList *n);
 void d_freefunc(GelEFunc *n);
 
 /*replace old with stuff from new and free new*/
-void d_replacefunc(GelEFunc *old,GelEFunc *new);
+void d_replacefunc (GelEFunc *old, GelEFunc *_new);
 
 /*push a new dictionary onto the context stack*/
-int d_addcontext(void);
+gboolean d_addcontext(void);
 
 /*gimme the last dictinary and pop the context stack*/
 void d_popcontext(void);
@@ -106,6 +104,10 @@ void d_popcontext(void);
 /*gimme the current dictinary*/
 GSList * d_getcontext(void);
 
+/*gimme the current global dictinary*/
+GSList * d_getcontext_global (void);
+GSList * d_find_similar_globals (const char *id);
+
 /* Put on subst local var list for this current stack */
 void d_put_on_subst_list (GelEFunc *func);
 
@@ -117,13 +119,15 @@ void d_protect_all(void);
 void d_add_named_args (GelEFunc *f, const char *args);
 
 #define D_ENSURE_USER_BODY(f) \
-	if (f->data.user == NULL) {					\
+	if G_UNLIKELY (f->data.user == NULL) {				\
 		g_assert (uncompiled != NULL);				\
 		f->data.user =						\
 			gel_decompile_tree (g_hash_table_lookup		\
 					    (uncompiled, f->id));	\
+		/* On error give null tree */				\
+		if (f->data.user == NULL)				\
+			f->data.user = gel_makenum_null ();		\
 		g_hash_table_remove (uncompiled, f->id);		\
-		g_assert (f->data.user != NULL);			\
 	}								\
 
 
