@@ -1,7 +1,7 @@
 /* GENIUS Calculator
- * Copyright (C) 1997-2004 George Lebl
+ * Copyright (C) 1997-2006 Jiri (George) Lebl
  *
- * Author: George Lebl
+ * Author: Jiri (George) Lebl
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -45,6 +45,8 @@
 #include "plugin.h"
 #include "inter.h"
 
+#include <vicious.h>
+
 #include <readline/readline.h>
 #include <readline/history.h>
 
@@ -55,15 +57,17 @@
 #include <gtksourceview/gtksourceprintjob.h>
 #endif
 
+#include <libgnomevfs/gnome-vfs-uri.h>
+#include <libgnomevfs/gnome-vfs-ops.h>
+#include <libgnomevfs/gnome-vfs-utils.h>
+
 #include "gnome-genius.h"
 
 /*Globals:*/
 
-#define DEFAULT_FONT "Monospace 10"
-
 /*calculator state*/
 calcstate_t curstate={
-	256,
+	128,
 	12,
 	FALSE,
 	FALSE,
@@ -75,6 +79,7 @@ calcstate_t curstate={
 	
 extern int parenth_depth;
 extern gboolean interrupted;
+extern const char *genius_toplevels[];
 
 GtkWidget *genius_window = NULL;
 
@@ -94,10 +99,12 @@ static int errors_printed = 0;
 static char *last_dir = NULL;
 
 GeniusSetup genius_setup = {
-	FALSE,
-	TRUE,
-	1000,
-	NULL
+	FALSE /* error_box */,
+	TRUE /* info_box */,
+	TRUE /* blinking_cursor */,
+	1000 /* scrollback */,
+	NULL /* font */,
+	FALSE /* black on white */
 };
 
 typedef struct {
@@ -113,17 +120,52 @@ typedef struct {
 	GtkWidget *label;
 } Program;
 
+enum {
+	TARGET_URI_LIST = 100
+};
+
+#define TERMINAL_PALETTE_SIZE 16
+const GdkColor
+terminal_palette_black_on_white[TERMINAL_PALETTE_SIZE] =
+{
+  { 0, 0x0000, 0x0000, 0x0000 },
+  { 0, 0xaaaa, 0x0000, 0x0000 },
+  { 0, 0x0000, 0x8888, 0x0000 },
+  { 0, 0xaaaa, 0x5555, 0x0000 },
+  { 0, 0x0000, 0x0000, 0xaaaa },
+  { 0, 0xaaaa, 0x0000, 0xaaaa },
+  { 0, 0x0000, 0xaaaa, 0xaaaa },
+  { 0, 0xaaaa, 0xaaaa, 0xaaaa },
+
+  { 0, 0x0000, 0x0000, 0x0000 },
+  { 0, 0xaaaa, 0x0000, 0x0000 },
+  { 0, 0x0000, 0x8888, 0x0000 },
+  { 0, 0xaaaa, 0x5555, 0x0000 },
+  { 0, 0x0000, 0x0000, 0xaaaa },
+  { 0, 0xaaaa, 0x0000, 0xaaaa },
+  { 0, 0x0000, 0x8888, 0xaaaa },
+  { 0, 0xaaaa, 0xaaaa, 0xaaaa },
+};
+
+static GtkTargetEntry drag_types[] = {
+	{ "text/uri-list", 0, TARGET_URI_LIST },
+};
+
+static gint n_drag_types = sizeof (drag_types) / sizeof (drag_types [0]);
+
 static Program *selected_program = NULL;
 static Program *running_program = NULL;
 
+static char *default_console_font = NULL;
+
 pid_t helper_pid = -1;
 
 static FILE *torlfp = NULL;
 static int fromrl;
 
-static int forzvt[2];
+static int forvte[2];
 
-static GIOChannel *forzvt0_ch;
+static GIOChannel *forvte0_ch;
 
 
 static char *torlfifo = NULL;
@@ -131,7 +173,7 @@ static char *fromrlfifo = NULL;
 
 static char *arg0 = NULL;
 
-static gboolean feed_to_zvt (GIOChannel *source, GIOCondition condition, gpointer data);
+static gboolean feed_to_vte (GIOChannel *source, GIOCondition condition, gpointer data);
 static void new_callback (GtkWidget *menu_item, gpointer data);
 static void open_callback (GtkWidget *w);
 static void save_callback (GtkWidget *w);
@@ -153,9 +195,11 @@ static void copy_as_troff (GtkWidget *menu_item, gpointer data);
 static void copy_as_mathml (GtkWidget *menu_item, gpointer data);
 static void setup_calc (GtkWidget *widget, gpointer data);
 static void run_program (GtkWidget *menu_item, gpointer data);
-static void manual_call (GtkWidget *widget, gpointer data);
 static void warranty_call (GtkWidget *widget, gpointer data);
 static void aboutcb (GtkWidget * widget, gpointer data);
+static void help_on_function (GtkWidget *menuitem, gpointer data);
+static void executing_warning (void);
+static void display_warning (GtkWidget *parent, const char *warn);
 
 static GnomeUIInfo file_menu[] = {
 	GNOMEUIINFO_MENU_NEW_ITEM(N_("_New Program"), N_("Create new program tab"), new_callback, NULL),
@@ -211,16 +255,15 @@ static GnomeUIInfo calc_menu[] = {
 	GNOMEUIINFO_ITEM_STOCK(N_("_Run"),N_("Run current program"),run_program, GTK_STOCK_EXECUTE),
 	GNOMEUIINFO_ITEM_STOCK(N_("_Interrupt"),N_("Interrupt current calculation"),genius_interrupt_calc,GTK_STOCK_STOP),
 	GNOMEUIINFO_SEPARATOR,
-	GNOMEUIINFO_ITEM_STOCK (N_("_Plot"), N_("Plot a function"), genius_lineplot_dialog, GNOME_STOCK_BOOK_OPEN),
+	GNOMEUIINFO_ITEM_STOCK (N_("_Plot"), N_("Plot a function"), genius_plot_dialog, GNOME_STOCK_BOOK_OPEN),
 	GNOMEUIINFO_END,
 };
 
 static GnomeUIInfo help_menu[] = {  
-	/* FIXME: no help 
-	 * GNOMEUIINFO_HELP("genius"),*/
-	GNOMEUIINFO_ITEM_STOCK (N_("_Manual"),
-				N_("Display the manual"),
-				manual_call,
+	GNOMEUIINFO_HELP("genius"),
+	GNOMEUIINFO_ITEM_STOCK (N_("_Help on Function"),
+				N_("Help on a function or a command"),
+				help_on_function,
 				GTK_STOCK_HELP),
 	GNOMEUIINFO_ITEM_STOCK (N_("_Warranty"),
 				N_("Display warranty information"),
@@ -256,8 +299,9 @@ static GnomeUIInfo toolbar[] = {
 	GNOMEUIINFO_ITEM_STOCK(N_("Interrupt"),N_("Interrupt current calculation"),genius_interrupt_calc,GTK_STOCK_STOP),
 #define TOOLBAR_RUN_ITEM 1
 	GNOMEUIINFO_ITEM_STOCK(N_("Run"),N_("Run current program"),run_program, GTK_STOCK_EXECUTE),
+	GNOMEUIINFO_ITEM_STOCK(N_("New"),N_("Create new program tab"), new_callback, GTK_STOCK_NEW),
 	GNOMEUIINFO_ITEM_STOCK(N_("Open"),N_("Open a GEL file for running"), open_callback, GTK_STOCK_OPEN),
-	GNOMEUIINFO_ITEM_STOCK(N_("Plot"), N_("Plot a function"), genius_lineplot_dialog, GNOME_STOCK_BOOK_OPEN),
+	GNOMEUIINFO_ITEM_STOCK(N_("Plot"), N_("Plot a function"), genius_plot_dialog, GNOME_STOCK_BOOK_OPEN),
 	GNOMEUIINFO_ITEM_STOCK(N_("Exit"),N_("Exit genius"), quitapp, GTK_STOCK_QUIT),
 	GNOMEUIINFO_END,
 };
@@ -265,6 +309,23 @@ static GnomeUIInfo toolbar[] = {
 
 #define ELEMENTS(x) (sizeof (x) / sizeof (x [0]))
 
+void
+genius_setup_window_cursor (GtkWidget *win, GdkCursorType type)
+{
+	GdkCursor *cursor = gdk_cursor_new (type);
+	if (win != NULL && win->window != NULL)
+		gdk_window_set_cursor (win->window, cursor);
+	gdk_cursor_unref (cursor);
+}
+
+void
+genius_unsetup_window_cursor (GtkWidget *win)
+{
+	if (win != NULL && win->window != NULL)
+		gdk_window_set_cursor (win->window, NULL);
+}
+
+
 static int
 count_char (const char *s, char c)
 {
@@ -276,6 +337,122 @@ count_char (const char *s, char c)
 	return i;
 }
 
+static gboolean
+uri_exists (const gchar* text_uri)
+{
+	GnomeVFSURI *uri;
+	gboolean res;
+		
+	g_return_val_if_fail (text_uri != NULL, FALSE);
+	
+	uri = gnome_vfs_uri_new (text_uri);
+	g_return_val_if_fail (uri != NULL, FALSE);
+
+	res = gnome_vfs_uri_exists (uri);
+
+	gnome_vfs_uri_unref (uri);
+
+	return res;
+}
+
+static void
+setup_term_color (void)
+{
+	if (genius_setup.black_on_white) {
+		GdkColor black = {0, 0, 0, 0};
+		GdkColor white = {0, 65535, 65535, 65535};
+		vte_terminal_set_colors (VTE_TERMINAL (term),
+					 &black,
+					 &white,
+					 terminal_palette_black_on_white,
+					 TERMINAL_PALETTE_SIZE);
+	} else {
+		vte_terminal_set_default_colors (VTE_TERMINAL (term));
+	}
+}
+
+static void
+help_on_entry_activate (GtkWidget *e, gpointer data)
+{
+	GtkWidget *d = data;
+	gtk_dialog_response (GTK_DIALOG (d), GTK_RESPONSE_OK);
+}
+
+static void
+help_on_function (GtkWidget *menuitem, gpointer data)
+{
+	GtkWidget *d;
+	GtkWidget *e;
+	int ret;
+
+	d = gtk_dialog_new_with_buttons
+		(_("Help on Function"),
+		 GTK_WINDOW (genius_window) /* parent */,
+		 0 /* flags */,
+		 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+		 GTK_STOCK_OK, GTK_RESPONSE_OK,
+		 NULL);
+
+	gtk_dialog_set_default_response (GTK_DIALOG (d), GTK_RESPONSE_OK);
+
+	gtk_dialog_set_has_separator (GTK_DIALOG (d), FALSE);
+	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d)->vbox),
+			    gtk_label_new (_("Function or command name:")),
+			    FALSE, FALSE, 0);
+
+	e = gtk_entry_new ();
+	g_signal_connect (G_OBJECT (e), "activate",
+			  G_CALLBACK (help_on_entry_activate), d);
+	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d)->vbox),
+			    e,
+			    FALSE, FALSE, 0);
+
+run_help_dlg_again:
+	gtk_widget_show_all (d);
+	ret = gtk_dialog_run (GTK_DIALOG (d));
+
+	if (ret == GTK_RESPONSE_OK) {
+		char *txt = g_strstrip (g_strdup (gtk_entry_get_text (GTK_ENTRY (e))));
+		GelHelp *help = get_help (txt, FALSE /* insert */);
+		gboolean found = FALSE;
+		int i;
+
+		for (i = 0; genius_toplevels[i] != NULL; i++) {
+			if (strcmp (txt, genius_toplevels[i]) == 0) {
+				gel_call_help (txt);
+				found = TRUE;
+				break;
+			}
+		}
+
+		if ( ! found && help == NULL) {
+			char *similar_ids = gel_similar_possible_ids (txt);
+			char *warn;
+			if (similar_ids == NULL) {
+				warn = g_strdup_printf
+					(_("<b>Help on %s not found</b>"),
+					 txt);
+			} else {
+				warn = g_strdup_printf
+					(_("<b>Help on %s not found</b>\n\n"
+					   "Perhaps you meant %s."),
+					 txt, similar_ids);
+				g_free (similar_ids);
+			}
+			display_warning (genius_window, warn);
+			g_free (warn);
+			goto run_help_dlg_again;
+		} else if ( ! found && help != NULL) {
+			if (help->aliasfor != NULL)
+				gel_call_help (help->aliasfor);
+			else
+				gel_call_help (txt);
+		}
+		g_free (txt);
+	}
+
+	gtk_widget_destroy (d);
+}
 
 /*display a message in a messagebox*/
 static GtkWidget *
@@ -284,7 +461,7 @@ geniusbox (gboolean error,
 	   const char *s)
 {
 	GtkWidget *mb;
-	/* if less then 10 lines */
+	/* if less than 10 lines */
 	if (count_char (s, '\n') <= 10) {
 		GtkMessageType type = GTK_MESSAGE_INFO;
 		if (error)
@@ -438,6 +615,34 @@ gel_printout_infos (void)
 	printout_error_num_and_reset ();
 }
 
+void
+gel_call_help (const char *function)
+{
+	if (function == NULL) {
+		/* FIXME: errors */
+
+		gnome_help_display ("genius", NULL, NULL /* error */);
+	} else {
+		char *id = NULL;
+		int i;
+		for (i = 0; genius_toplevels[i] != NULL && id == NULL; i++) {
+			if (strcmp (function, genius_toplevels[i]) == 0) {
+				id = g_strdup_printf ("gel-command-%s",
+						      function);
+				break;
+			}
+		}
+		if (id == NULL) {
+			id = g_strdup_printf ("gel-function-%s",
+					      function);
+		}
+
+		/* FIXME: errors */
+
+		gnome_help_display ("genius", id, NULL /* error */);
+	}
+}
+
 
 /*get info message*/
 static void
@@ -490,7 +695,23 @@ aboutcb(GtkWidget * widget, gpointer data)
 		 * which will give them credit in the About box.
 		 * E.g. "Fulano de Tal <fulano@detal.com>"
 		 */
-		translators = _("translator_credits-PLEASE_ADD_YOURSELF_HERE");
+		char *new_credits = N_("translator-credits");
+
+		/* hack for old translations */
+		char *old_hack = "translator_credits-PLEASE_ADD_YOURSELF_HERE";
+
+		translators = _(new_credits);
+		if (strcmp (translators, new_credits) == 0) {
+			translators = NULL;
+		}
+
+		/* hack for old translations */
+		if (translators == NULL) {
+			translators = _(old_hack);
+			if (strcmp (translators, old_hack) == 0) {
+				translators = NULL;
+			}
+		}
 
 		about = gnome_about_new
 			(_("About Genius"),
@@ -501,9 +722,8 @@ aboutcb(GtkWidget * widget, gpointer data)
 			   "details, type 'warranty' into the console."),
 			 authors,
 			 documenters,
-			(strcmp (translators, "translator_credits-PLEASE_ADD_YOURSELF_HERE")
-			 ? translators : NULL),
-			NULL);
+			 translators,
+			 NULL);
 
 		gtk_window_set_transient_for (GTK_WINDOW (about),
 					      GTK_WINDOW (genius_window));
@@ -520,12 +740,18 @@ aboutcb(GtkWidget * widget, gpointer data)
 static void
 set_properties (void)
 {
-	gnome_config_set_string("/genius/properties/pango_font", genius_setup.font?
-				genius_setup.font:DEFAULT_FONT);
-	gnome_config_set_int("/genius/properties/scrollback", genius_setup.scrollback);
-	gnome_config_set_bool("/genius/properties/error_box", genius_setup.error_box);
+	gnome_config_set_bool("/genius/properties/black_on_white",
+			      genius_setup.black_on_white);
+	gnome_config_set_string ("/genius/properties/pango_font",
+				 ve_sure_string (genius_setup.font));
+	gnome_config_set_int("/genius/properties/scrollback",
+			     genius_setup.scrollback);
+	gnome_config_set_bool("/genius/properties/error_box",
+			      genius_setup.error_box);
 	gnome_config_set_bool("/genius/properties/info_box",
 			      genius_setup.info_box);
+	gnome_config_set_bool("/genius/properties/blinking_cursor",
+			      genius_setup.blinking_cursor);
 	gnome_config_set_int("/genius/properties/max_digits", 
 			      curstate.max_digits);
 	gnome_config_set_bool("/genius/properties/results_as_floats",
@@ -536,12 +762,14 @@ set_properties (void)
 			      curstate.full_expressions);
 	gnome_config_set_int("/genius/properties/max_errors",
 			     curstate.max_errors);
+	gnome_config_set_int("/genius/properties/float_prec",
+			     curstate.float_prec);
 	
 	gnome_config_sync();
 }
 
-static void
-display_error (GtkWidget *parent, const char *err)
+void
+genius_display_error (GtkWidget *parent, const char *err)
 {
 	static GtkWidget *w = NULL;
 
@@ -596,8 +824,8 @@ display_warning (GtkWidget *parent, const char *warn)
 	gtk_widget_destroy (w);
 }
 
-static gboolean
-ask_question (GtkWidget *parent, const char *question)
+gboolean
+genius_ask_question (GtkWidget *parent, const char *question)
 {
 	int ret;
 	static GtkWidget *req = NULL;
@@ -657,31 +885,31 @@ quitapp (GtkWidget * widget, gpointer data)
 {
 	if (any_changed ()) {
 		if (calc_running) {
-			if ( ! ask_question (NULL,
-					     _("Genius is executing something, "
-					       "and furthermore there are "
-					       "unsaved programs.\nAre "
-					       "you sure you wish to quit?")))
+			if ( ! genius_ask_question (NULL,
+						    _("Genius is executing something, "
+						      "and furthermore there are "
+						      "unsaved programs.\nAre "
+						      "you sure you wish to quit?")))
 				return;
 			interrupted = TRUE;
 		} else {
-			if ( ! ask_question (NULL,
-					     _("There are unsaved programs, "
-					       "are you sure you wish to quit?")))
+			if ( ! genius_ask_question (NULL,
+						    _("There are unsaved programs, "
+						      "are you sure you wish to quit?")))
 				return;
 		}
 	} else {
 		if (calc_running) {
-			if ( ! ask_question (NULL,
-					     _("Genius is executing something, "
-					       "are you sure you wish to "
-					       "quit?")))
+			if ( ! genius_ask_question (NULL,
+						    _("Genius is executing something, "
+						      "are you sure you wish to "
+						      "quit?")))
 				return;
 			interrupted = TRUE;
 		} else {
-			if ( ! ask_question (NULL,
-					     _("Are you sure you wish "
-					       "to quit?")))
+			if ( ! genius_ask_question (NULL,
+						    _("Are you sure you wish "
+						      "to quit?")))
 				return;
 		}
 	}
@@ -743,9 +971,17 @@ setup_response (GtkWidget *widget, gint resp, gpointer data)
 		set_new_calcstate (curstate);
 		vte_terminal_set_scrollback_lines (VTE_TERMINAL (term),
 						   genius_setup.scrollback);
-		vte_terminal_set_font_from_string (VTE_TERMINAL (term),
-						   genius_setup.font ?
-						   genius_setup.font : DEFAULT_FONT);
+		vte_terminal_set_font_from_string
+			(VTE_TERMINAL (term),
+			 ve_string_empty (genius_setup.font) ?
+			   default_console_font :
+			   genius_setup.font);
+		setup_term_color ();
+		vte_terminal_set_cursor_blinks
+			(VTE_TERMINAL (term),
+			genius_setup.blinking_cursor);
+
+
 		if (resp == GTK_RESPONSE_OK ||
 		    resp == GTK_RESPONSE_CANCEL)
 			gtk_widget_destroy (widget);
@@ -988,8 +1224,9 @@ setup_calc(GtkWidget *widget, gpointer data)
 	
         w = gnome_font_picker_new();
 	gnome_font_picker_set_font_name (GNOME_FONT_PICKER (w),
-					 tmpsetup.font ? tmpsetup.font :
-					 DEFAULT_FONT);
+					 ve_string_empty (tmpsetup.font) ?
+					   default_console_font :
+					   genius_setup.font);
         gnome_font_picker_set_mode (GNOME_FONT_PICKER (w),
 				    GNOME_FONT_PICKER_MODE_FONT_INFO);
         gtk_box_pack_start(GTK_BOX(b),w,TRUE,TRUE,0);
@@ -997,6 +1234,22 @@ setup_calc(GtkWidget *widget, gpointer data)
 			  G_CALLBACK (fontsetcb),
 			  &tmpsetup.font);
 
+	w=gtk_check_button_new_with_label(_("Black on white"));
+	gtk_box_pack_start(GTK_BOX(box),w,FALSE,FALSE,0);
+	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w), 
+				      tmpsetup.black_on_white);
+	g_signal_connect (G_OBJECT(w), "toggled",
+			  G_CALLBACK (optioncb),
+			  (gpointer)&tmpsetup.black_on_white);
+
+	w=gtk_check_button_new_with_label(_("Blinking cursor"));
+	gtk_box_pack_start(GTK_BOX(box),w,FALSE,FALSE,0);
+	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w), 
+				      tmpsetup.blinking_cursor);
+	g_signal_connect (G_OBJECT(w), "toggled",
+			  G_CALLBACK (optioncb),
+			  (gpointer)&tmpsetup.blinking_cursor);
+
 
 	g_signal_connect (G_OBJECT (setupdialog), "response",
 			  G_CALLBACK (setup_response), NULL);	
@@ -1024,22 +1277,6 @@ executing_warning (void)
 			   "operation."));
 }
 
-static void
-manual_call (GtkWidget *widget, gpointer data)
-{
-	if (calc_running) {
-		executing_warning ();
-		return;
-	} else {
-		/* perhaps a bit ugly */
-		gboolean last = genius_setup.info_box;
-		genius_setup.info_box = TRUE;
-		gel_evalexp ("manual", NULL, main_out, NULL, TRUE, NULL);
-		gel_printout_infos ();
-		genius_setup.info_box = last;
-	}
-}
-
 static void
 warranty_call (GtkWidget *widget, gpointer data)
 {
@@ -1121,9 +1358,9 @@ really_load_cb (GtkWidget *w, GtkFileSelection *fs)
 	s = gtk_file_selection_get_filename (fs);
 #endif
 	if (s == NULL ||
-	    access (s, R_OK) != 0) {
-		display_error (GTK_WIDGET (fs),
-			       _("Cannot open file!"));
+	    access (s, F_OK) != 0) {
+		genius_display_error (GTK_WIDGET (fs),
+				      _("Cannot open file!"));
 		return;
 	}
 
@@ -1478,12 +1715,86 @@ changed_cb (GtkTextBuffer *buffer, GtkWidget *tab_widget)
 	}
 }
 
+static gboolean
+save_contents_vfs (const char *file, const char *str, int size)
+{
+	GnomeVFSHandle *handle;
+	GnomeVFSFileSize bytes;
+	GnomeVFSResult result;
+
+	/* FIXME: we should handle errors better by perhaps moving
+	   to a different name first and erasing only when saving
+	   was all fine */
+
+	/* Be safe about saving files, unlink and create in
+	 * exclusive mode */
+	result = gnome_vfs_unlink (file);
+	/* FIXME: error handling, but not if it's
+	 * the file-doesn't-exist kind of error which is fine */
+	result = gnome_vfs_create (&handle, file,
+				   GNOME_VFS_OPEN_WRITE,
+				   TRUE /* exclusive */,
+				   0644);
+	if (result != GNOME_VFS_OK) {
+		/* FIXME: error handling */
+		return FALSE;
+	}
+
+	result = gnome_vfs_write (handle, str, size, &bytes);
+	if (result != GNOME_VFS_OK || bytes != size) {
+		gnome_vfs_close (handle);
+		/* FIXME: error handling */
+		return FALSE;
+	}
+
+	/* add traling \n if needed */
+	if (size > 0 && str[size-1] != '\n')
+		gnome_vfs_write (handle, "\n", 1, &bytes);
+	/* FIXME: error handling? */
+
+	gnome_vfs_close (handle);
+
+	return TRUE;
+}
+
+static char *
+get_contents_vfs (const char *file)
+{
+	GnomeVFSHandle *handle;
+	GnomeVFSFileSize bytes;
+	char buffer[4096];
+	GnomeVFSResult result;
+	GString *str;
+
+	/* FIXME: add limit to avoid reading until never */
+
+	result = gnome_vfs_open (&handle, file,
+				 GNOME_VFS_OPEN_READ);
+	if (result != GNOME_VFS_OK) {
+		/* FIXME: error handling */
+		return NULL;
+	}
+
+	str = g_string_new (NULL);
+
+	while (gnome_vfs_read (handle,
+			       buffer,
+			       sizeof (buffer)-1,
+			       &bytes) == GNOME_VFS_OK) {
+		buffer[bytes] = '\0';
+		g_string_append (str, buffer);
+	}
+
+	gnome_vfs_close (handle);
+
+	return g_string_free (str, FALSE);
+}
+
 static void
 reload_cb (GtkWidget *menu_item)
 {
 	GtkTextIter iter, iter_end;
 	char *contents;
-	int len;
 
 	if (selected_program == NULL ||
 	    ! selected_program->real_file)
@@ -1498,19 +1809,17 @@ reload_cb (GtkWidget *menu_item)
 	gtk_text_buffer_delete (selected_program->buffer,
 				&iter, &iter_end);
 
-	if (g_file_get_contents (selected_program->name,
-				 &contents,
-				 &len,
-				 NULL /* FIXME: error */)) {
+	contents = get_contents_vfs (selected_program->name);
+	if (contents != NULL) {
 		gtk_text_buffer_get_iter_at_offset (selected_program->buffer,
 						    &iter, 0);
 		gtk_text_buffer_insert_with_tags_by_name
 			(selected_program->buffer,
-			 &iter, contents, len, "foo", NULL);
+			 &iter, contents, -1, "foo", NULL);
 		g_free (contents);
 		selected_program->changed = FALSE;
 	} else {
-		display_error (NULL, _("Cannot open file"));
+		genius_display_error (NULL, _("Cannot open file"));
 	}
 
 	selected_program->ignore_changes--;
@@ -1612,26 +1921,27 @@ new_program (const char *filename)
 				p);
 
 	if (filename == NULL) {
+		char *d = g_get_current_dir ();
+		char *n = g_strdup_printf (_("Program_%d.gel"), cnt);
 		/* the file name will have an underscore */
-		p->name = g_strdup_printf (_("Program_%d.gel"), cnt);
+		p->name = g_strconcat (d, "/", n, NULL);
+		g_free (d);
+		g_free (n);
 		p->vname = g_strdup_printf (_("Program %d"), cnt);
 		cnt++;
 	} else {
 		char *contents;
-		int len;
 		p->name = g_strdup (filename);
-		if (g_file_get_contents (filename,
-					 &contents,
-					 &len,
-					 NULL /* FIXME: error */)) {
+		contents = get_contents_vfs (filename);
+		if (contents != NULL) {
 			GtkTextIter iter;
 			gtk_text_buffer_get_iter_at_offset (buffer, &iter, 0);
 			gtk_text_buffer_insert_with_tags_by_name
-				(buffer, &iter, contents, len, "foo", NULL);
+				(buffer, &iter, contents, -1, "foo", NULL);
 			g_free (contents);
 		} else {
 			char *s = g_strdup_printf (_("Cannot open %s"), filename);
-			display_error (NULL, s);
+			genius_display_error (NULL, s);
 			g_free (s);
 		}
 		p->vname = g_path_get_basename (p->name);
@@ -1671,14 +1981,14 @@ really_open_cb (GtkWidget *w, GtkFileSelection *fs)
 #endif
 
 #if GTK_CHECK_VERSION(2,3,5)
-	s = gtk_file_chooser_get_filename (fs);
+	s = gtk_file_chooser_get_uri (fs);
 #else
 	s = gtk_file_selection_get_filename (fs);
 #endif
 	if (s == NULL ||
-	    access (s, R_OK) != 0) {
-		display_error (GTK_WIDGET (fs),
-			       _("Cannot open file!"));
+	    ! uri_exists (s)) {
+		genius_display_error (GTK_WIDGET (fs),
+				      _("Cannot open file!"));
 		return;
 	}
 
@@ -1711,6 +2021,7 @@ open_callback (GtkWidget *w)
 					  GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
 					  GTK_STOCK_OPEN, GTK_RESPONSE_OK,
 					  NULL);
+	gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (fs), FALSE);
 
 	add_filters (GTK_FILE_CHOOSER (fs));
 
@@ -1748,42 +2059,27 @@ open_callback (GtkWidget *w)
 static gboolean
 save_program (Program *p, const char *new_fname)
 {
-	FILE *fp;
 	GtkTextIter iter, iter_end;
 	char *prog;
 	const char *fname;
 	int sz;
-	int wrote;
 
 	if (new_fname == NULL)
 		fname = p->name;
 	else
 		fname = new_fname;
 
-	fp = fopen (fname, "w");
-	if (fp == NULL)
-		return FALSE;
-
 	gtk_text_buffer_get_iter_at_offset (p->buffer, &iter, 0);
 	gtk_text_buffer_get_iter_at_offset (p->buffer, &iter_end, -1);
 	prog = gtk_text_buffer_get_text (p->buffer, &iter, &iter_end,
 					 FALSE /* include_hidden_chars */);
 	sz = strlen (prog);
 
-	wrote = fwrite (prog, 1, sz, fp);
-	if (sz != wrote) {
+	if ( ! save_contents_vfs (fname, prog, sz)) {
 		g_free (prog);
-		fclose (fp);
 		return FALSE;
 	}
 
-	/* add trailing \n */
-	if (sz > 0 && prog[sz-1] != '\n')
-		fwrite ("\n", 1, 1, fp);
-
-	g_free (prog);
-	fclose (fp);
-
 	if (p->name != fname) {
 		g_free (p->name);
 		p->name = g_strdup (fname);
@@ -1816,7 +2112,7 @@ save_callback (GtkWidget *w)
 		char *err = g_strdup_printf (_("<b>Cannot save file</b>\n"
 					       "Details: %s"),
 					     g_strerror (errno));
-		display_error (NULL, err);
+		genius_display_error (NULL, err);
 		g_free (err);
 	}
 }
@@ -1845,7 +2141,7 @@ really_save_as_cb (GtkWidget *w, GtkFileSelection *fs)
 		return;
 
 #if GTK_CHECK_VERSION(2,3,5)
-	s = g_strdup (gtk_file_chooser_get_filename (fs));
+	s = g_strdup (gtk_file_chooser_get_uri (fs));
 #else
 	s = g_strdup (gtk_file_selection_get_filename (fs));
 #endif
@@ -1860,8 +2156,8 @@ really_save_as_cb (GtkWidget *w, GtkFileSelection *fs)
 	}
 	g_free (base);
 	
-	if (access (s, F_OK) == 0 &&
-	    ! ask_question (GTK_WIDGET (fs),
+	if (uri_exists (s) &&
+	    ! genius_ask_question (GTK_WIDGET (fs),
 			    _("File already exists.  Overwrite it?"))) {
 		g_free (s);
 		return;
@@ -1871,7 +2167,7 @@ really_save_as_cb (GtkWidget *w, GtkFileSelection *fs)
 		char *err = g_strdup_printf (_("<b>Cannot save file</b>\n"
 					       "Details: %s"),
 					     g_strerror (errno));
-		display_error (GTK_WIDGET (fs), err);
+		genius_display_error (GTK_WIDGET (fs), err);
 		g_free (err);
 		g_free (s);
 		return;
@@ -1925,6 +2221,7 @@ save_as_callback (GtkWidget *w)
 					  GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
 					  GTK_STOCK_SAVE, GTK_RESPONSE_OK,
 					  NULL);
+	gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (fs), FALSE);
 
 	add_filters (GTK_FILE_CHOOSER (fs));
 
@@ -2013,10 +2310,10 @@ run_program (GtkWidget *menu_item, gpointer data)
 	const char *name;
 	GtkTextBuffer *buffer;
 	if (selected_program == NULL) /* if nothing is selected */ {
-		display_error (NULL,
-			       _("<b>No program selected.</b>\n\n"
-				 "Create a new program, or select an "
-				 "existing tab in the notebook."));
+		genius_display_error (NULL,
+				      _("<b>No program selected.</b>\n\n"
+					"Create a new program, or select an "
+					"existing tab in the notebook."));
 		return;
 	}
 	buffer = selected_program->buffer;
@@ -2106,6 +2403,14 @@ run_program (GtkWidget *menu_item, gpointer data)
 
 }
 
+static gboolean
+delete_event (GtkWidget *w, GdkEventAny *e, gpointer data)
+{
+	quitapp (w, data);
+	return TRUE;
+}
+
+
 /*main window creation, slightly copied from same-gnome:)*/
 static GtkWidget *
 create_main_window (void)
@@ -2117,7 +2422,7 @@ create_main_window (void)
 	gtk_window_set_wmclass (GTK_WINDOW (w), "gnome-genius", "gnome-genius");
 
         g_signal_connect (G_OBJECT (w), "delete_event",
-			  G_CALLBACK (quitapp), NULL);
+			  G_CALLBACK (delete_event), NULL);
         return w;
 }
 
@@ -2128,38 +2433,67 @@ get_properties (void)
 {
 	gchar buf[256];
 
+	g_snprintf(buf,256,"/genius/properties/black_on_white=%s",
+		   (genius_setup.black_on_white)?"true":"false");
+	genius_setup.black_on_white = gnome_config_get_bool(buf);
+
 	g_snprintf (buf, 256, "/genius/properties/pango_font=%s",
-		    genius_setup.font ? genius_setup.font : DEFAULT_FONT);
+		    ve_sure_string (genius_setup.font));
 	genius_setup.font = gnome_config_get_string (buf);
+
 	g_snprintf(buf,256,"/genius/properties/scrollback=%d",
 		   genius_setup.scrollback);
 	genius_setup.scrollback = gnome_config_get_int(buf);
+
 	g_snprintf(buf,256,"/genius/properties/error_box=%s",
 		   (genius_setup.error_box)?"true":"false");
 	genius_setup.error_box = gnome_config_get_bool(buf);
+
 	g_snprintf(buf,256,"/genius/properties/info_box=%s",
 		   (genius_setup.info_box)?"true":"false");
 	genius_setup.info_box = gnome_config_get_bool(buf);
+
+	g_snprintf(buf,256,"/genius/properties/blinking_cursor=%s",
+		   (genius_setup.blinking_cursor)?"true":"false");
+	genius_setup.blinking_cursor = gnome_config_get_bool(buf);
 	
 	g_snprintf(buf,256,"/genius/properties/max_digits=%d",
 		   curstate.max_digits);
 	curstate.max_digits = gnome_config_get_int(buf);
+	if (curstate.max_digits < 0)
+		curstate.max_digits = 0;
+	else if (curstate.max_digits > 256)
+		curstate.max_digits = 256;
+
 	g_snprintf(buf,256,"/genius/properties/results_as_floats=%s",
 		   curstate.results_as_floats?"true":"false");
 	curstate.results_as_floats = gnome_config_get_bool(buf);
+
 	g_snprintf(buf,256,"/genius/properties/scientific_notation=%s",
 		   curstate.scientific_notation?"true":"false");
 	curstate.scientific_notation = gnome_config_get_bool(buf);
+
 	g_snprintf(buf,256,"/genius/properties/full_expressions=%s",
 		   curstate.full_expressions?"true":"false");
 	curstate.full_expressions = gnome_config_get_bool(buf);
+
 	g_snprintf(buf,256,"/genius/properties/max_errors=%d",
 		   curstate.max_errors);
 	curstate.max_errors = gnome_config_get_int(buf);
+	if (curstate.max_errors < 0)
+		curstate.float_prec = 0;
+
+	g_snprintf(buf,256,"/genius/properties/float_prec=%d",
+		   curstate.float_prec);
+	curstate.float_prec = gnome_config_get_int(buf);
+	if (curstate.float_prec < 60)
+		curstate.float_prec = 60;
+	else if (curstate.float_prec > 16384)
+		curstate.float_prec = 16384;
 }
 
 static void
-feed_to_zvt_from_string (const char *str, int size)
+feed_to_vte_from_string (const char *str, int size)
 {
 	/*do our own crlf translation*/
 	char *s;
@@ -2178,12 +2512,13 @@ feed_to_zvt_from_string (const char *str, int size)
 			s[sz] = '\r';
 		} else s[sz] = str[i];
 	}
-	vte_terminal_feed (VTE_TERMINAL (term), s, sz);
+	vte_terminal_feed (VTE_TERMINAL (term), 
+			   s, sz);
 	g_free(s);
 }
 
 static gboolean
-feed_to_zvt (GIOChannel *source, GIOCondition condition, gpointer data)
+feed_to_vte (GIOChannel *source, GIOCondition condition, gpointer data)
 {
 	if (condition & G_IO_IN) {
 		int fd = g_io_channel_unix_get_fd (source);
@@ -2192,7 +2527,7 @@ feed_to_zvt (GIOChannel *source, GIOCondition condition, gpointer data)
 		while ((size = read (fd, buf, 256)) > 0 ||
 		       errno == EINTR) {
 			if (size > 0)
-				feed_to_zvt_from_string (buf, size);
+				feed_to_vte_from_string (buf, size);
 		}
 	}
 
@@ -2204,8 +2539,8 @@ output_notify_func (GelOutput *output)
 {
 	const char *s = gel_output_peek_string (output);
 	if (s != NULL) {
-		feed_to_zvt (forzvt0_ch, G_IO_IN, NULL);
-		feed_to_zvt_from_string ((char *)s, strlen (s));
+		feed_to_vte (forvte0_ch, G_IO_IN, NULL);
+		feed_to_vte_from_string ((char *)s, strlen (s));
 		gel_output_clear_string (output);
 	}
 }
@@ -2387,12 +2722,25 @@ genius_got_etree (GelETree *e)
 }
 
 static char *
-make_a_fifo (void)
+make_a_fifo (const char *postfix)
 {
-	static int cnt = 1;
+	if (g_get_home_dir () != NULL) {
+		char *name = g_strdup_printf ("%s/.genius-fifo-%s",
+					      g_get_home_dir (),
+					      postfix);
+		/* this will not work if we don't own this, but this will
+		 * make sure we clean up old links */
+		unlink (name);
+		if (mkfifo (name, 0600) == 0) {
+			return name;
+		}
+		g_free (name);
+	}
+       
 	for (;;) {
-		char *name = g_strdup_printf ("/tmp/genius-fifo-%d-%d",
-					      (int)getpid(), cnt++);
+		char *name = g_strdup_printf ("/tmp/genius-fifo-%x-%s",
+					      (guint)g_random_int (),
+					      postfix);
 		/* this will not work if we don't own this, but this will
 		 * make sure we clean up old links */
 		unlink (name);
@@ -2406,8 +2754,8 @@ make_a_fifo (void)
 static void
 setup_rl_fifos (void)
 {
-	torlfifo = make_a_fifo ();
-	fromrlfifo = make_a_fifo ();
+	torlfifo = make_a_fifo ("torl");
+	fromrlfifo = make_a_fifo ("fromrl");
 }
 
 static void
@@ -2500,10 +2848,7 @@ get_version_details (void)
 	if (str != NULL)
 		return str->str;
 	str = g_string_new (NULL);
-#ifndef HAVE_MPFR
-	g_string_append (str, _("\nNote: Compiled without MPFR (some operations may be slow) "
-				"see www.mpfr.org"));
-#endif
+
 #ifndef HAVE_GTKSOURCEVIEW
 	g_string_append (str, _("\nNote: Compiled without GtkSourceView (better source editor)"));
 #endif
@@ -2529,6 +2874,63 @@ loadup_files_from_cmdline (GnomeProgram *program)
 	}
 }
 
+static void 
+drag_data_received (GtkWidget *widget, GdkDragContext *context, 
+		    gint x, gint y, GtkSelectionData *selection_data, 
+		    guint info, guint time)
+{
+	GList *list;
+	GList *li;
+	
+	if (info != TARGET_URI_LIST)
+		return;
+			
+	list = gnome_vfs_uri_list_parse ((gpointer)selection_data->data);
+
+	for (li = list; li != NULL; li = li->next) {
+		const GnomeVFSURI *uri = li->data;
+		char *s = gnome_vfs_uri_to_string (uri,
+						   GNOME_VFS_URI_HIDE_NONE);
+		new_program (s);
+	}
+	
+	gnome_vfs_uri_list_free (list);
+}
+
+static void
+update_term_geometry (void)
+{
+	GdkGeometry hints;
+	int char_width;
+	int char_height;
+	int xpad, ypad;
+
+	char_width = VTE_TERMINAL (term)->char_width;
+	char_height = VTE_TERMINAL (term)->char_height;
+  
+	vte_terminal_get_padding (VTE_TERMINAL (term), &xpad, &ypad);
+
+	hints.base_width = xpad;
+	hints.base_height = ypad;
+
+#define MIN_WIDTH_CHARS 10
+#define MIN_HEIGHT_CHARS 4
+
+	hints.width_inc = char_width;
+	hints.height_inc = char_height;
+
+	/* min size is min size of just the geometry widget, remember. */
+	hints.min_width = hints.base_width + hints.width_inc * MIN_WIDTH_CHARS;
+	hints.min_height = hints.base_height + hints.height_inc * MIN_HEIGHT_CHARS;
+
+	gtk_window_set_geometry_hints (GTK_WINDOW (genius_window),
+				       term,
+				       &hints,
+				       GDK_HINT_RESIZE_INC |
+				       GDK_HINT_MIN_SIZE |
+				       GDK_HINT_BASE_SIZE);
+}
+
 
 int
 main (int argc, char *argv[])
@@ -2553,19 +2955,20 @@ main (int argc, char *argv[])
 	program = gnome_program_init ("genius", VERSION, 
 				      LIBGNOMEUI_MODULE /* module_info */,
 				      argc, argv,
+				      GNOME_PARAM_APP_DATADIR, DATADIR,
 				      /* GNOME_PARAM_POPT_TABLE, options, */
 				      NULL);
 
-	if (pipe (forzvt) < 0)
+	if (pipe (forvte) < 0)
 		g_error ("Can't pipe");
 
 	setup_rl_fifos ();
 
-	fcntl (forzvt[0], F_SETFL, O_NONBLOCK);
+	fcntl (forvte[0], F_SETFL, O_NONBLOCK);
 
-	forzvt0_ch = g_io_channel_unix_new (forzvt[0]);
-	g_io_add_watch_full (forzvt0_ch, G_PRIORITY_DEFAULT, G_IO_IN | G_IO_HUP | G_IO_ERR, 
-			     feed_to_zvt, NULL, NULL);
+	forvte0_ch = g_io_channel_unix_new (forvte[0]);
+	g_io_add_watch_full (forvte0_ch, G_PRIORITY_DEFAULT, G_IO_IN | G_IO_HUP | G_IO_ERR, 
+			     feed_to_vte, NULL, NULL);
 
 	main_out = gel_output_new();
 	gel_output_setup_string (main_out, 80, get_term_width);
@@ -2582,6 +2985,18 @@ main (int argc, char *argv[])
         /*set up the top level window*/
 	genius_window = create_main_window();
 
+	/* Drag and drop support */
+	gtk_drag_dest_set (GTK_WIDGET (genius_window),
+			   GTK_DEST_DEFAULT_MOTION |
+			   GTK_DEST_DEFAULT_HIGHLIGHT |
+			   GTK_DEST_DEFAULT_DROP,
+			   drag_types, n_drag_types,
+			   GDK_ACTION_COPY);
+		
+	g_signal_connect (G_OBJECT (genius_window), "drag_data_received",
+			  G_CALLBACK (drag_data_received), 
+			  NULL);
+
 	/*set up the tooltips*/
 	tips = gtk_tooltips_new();
 
@@ -2609,7 +3024,6 @@ main (int argc, char *argv[])
 	/* FIXME: how come does backspace and not delete */
 	vte_terminal_set_delete_binding (VTE_TERMINAL (term),
 					 VTE_ERASE_ASCII_DELETE);
-
 	g_signal_connect (G_OBJECT (term), "selection_changed",
 			  G_CALLBACK (selection_changed),
 			  NULL);
@@ -2672,7 +3086,7 @@ main (int argc, char *argv[])
 				  hbox,
 				  gtk_label_new (_("Console")));
 	/* FIXME:
-	gtk_widget_queue_resize (zvt);
+	gtk_widget_queue_resize (vte);
 	*/
 	gtk_container_set_border_width(
 		GTK_CONTAINER (GNOME_APP (genius_window)->contents), 5);
@@ -2691,24 +3105,52 @@ main (int argc, char *argv[])
 
 	gtk_widget_show_all (genius_window);
 
+	/* Try to deduce the standard font size, kind of evil, but sorta
+	 * works.  The user can always set the font themselves. */
+	{
+		GtkStyle *style = gtk_widget_get_style (genius_window);
+		int sz = (style == NULL ||
+			  style->font_desc == NULL) ? 10 :
+			pango_font_description_get_size (style->font_desc) / PANGO_SCALE;
+		if (sz == 0) sz = 10;
+		default_console_font = g_strdup_printf ("Monospace %d", sz);
+	}
+
 	/* for some reason we must set the font here and not above
 	 * or the "monospace 12" (or default terminal font or whatnot)
 	 * will get used */
-	vte_terminal_set_font_from_string (VTE_TERMINAL (term), 
-					   genius_setup.font ?
-					   genius_setup.font : DEFAULT_FONT);
+	vte_terminal_set_font_from_string (VTE_TERMINAL (term),
+					   ve_string_empty (genius_setup.font) ?
+					     default_console_font :
+					     genius_setup.font);
+	setup_term_color ();
+	vte_terminal_set_cursor_blinks
+		(VTE_TERMINAL (term),
+		 genius_setup.blinking_cursor);
+	vte_terminal_set_encoding (VTE_TERMINAL (term), "UTF-8");
 
+	update_term_geometry ();
+	g_signal_connect (G_OBJECT (term), "char-size-changed",
+			  G_CALLBACK (update_term_geometry), NULL);
 
 	gtk_widget_show_now (genius_window);
 
 	gel_output_printf (main_out,
-			   _("\e[0;32mGenius %s\e[0m\n"
+			   _("%sGenius %s%s\n"
 			     "%s\n"
 			     "This is free software with ABSOLUTELY NO WARRANTY.\n"
-			     "For license details type `\e[01;36mwarranty\e[0m'.\n"
-			     "For help type '\e[01;36mmanual\e[0m' or '\e[01;36mhelp\e[0m'.%s\n\n"),
+			     "For license details type `%swarranty%s'.\n"
+			     "For help type '%smanual%s' or '%shelp%s'.%s\n\n"),
+			   "\e[0;32m" /* green */,
+			   "\e[0m" /* white on black */,
 			   VERSION,
 			   COPYRIGHT_STRING,
+			   "\e[01;36m" /* cyan */,
+			   "\e[0m" /* white on black */,
+			   "\e[01;36m" /* cyan */,
+			   "\e[0m" /* white on black */,
+			   "\e[01;36m" /* cyan */,
+			   "\e[0m" /* white on black */,
 			   get_version_details ());
 	gel_output_flush (main_out);
 	check_events ();
